import cors from "@fastify/cors";
import { Type, TypeBoxTypeProvider } from "@fastify/type-provider-typebox";
import Fastify, { FastifyReply, FastifyRequest } from "fastify";
import { nanoid } from "nanoid";
import { allocateBuffer, IMAGES } from "runner";
import { Pull, Push } from "zeromq";

import {
  createDbDirectory,
  openDatabase,
  closeDatabase,
  createTables,
  insertUser,
  updateUserLogin,
  updateUserPassword,
  updateUserPermissions,
  deleteUserById,
  deleteUserByLogin,
  deleteAllUsers,
  selectAllUsers,
  selectUserByLogin,
  selectUserById,
  insertLanguage,
  updateLanguageDesignation,
  updateLanguageVersion,
  deleteLanguage,
  deleteAllLanguages,
  selectAllLanguages,
  selectLanguageById,
  insertWork,
  selectAllWorks,
  deleteAllWorks,
  deleteWork,
  selectWorkById,
} from "./database";

const sender = new Push();
await sender.bind(`tcp://127.0.0.1:5557`);
const receiver = new Pull();
await receiver.bind(`tcp://127.0.0.1:5558`);

const clients: Record<string, FastifyReply> = {};
const generateId = () => nanoid(32);

const fastify = Fastify({
  logger: true,
}).withTypeProvider<TypeBoxTypeProvider>();
await fastify.register(cors, {
  origin: process.env.ALLOW_ORIGIN || "*",
});

/* Code runner in a container */

/* POST /run : Run code in a container */

fastify.post(
  "/run",
  {
    schema: {
      body: Type.Object({
        code: Type.String(),
        language: Type.String(),
      }),
    },
  },
  (req, reply) => {
    const { code, language } = req.body;
    const jobId = generateId();
    const buffer = allocateBuffer(jobId, code, IMAGES.moshell);
    reply.raw.writeHead(200, {
      "Content-Type": "text/event-stream",
      Connection: "keep-alive",
      "Cache-Control": "no-cache",
      "Access-Control-Allow-Origin": process.env.ALLOW_ORIGIN || "*",
    });
    reply.raw.on("close", () => {
      delete clients[jobId];
    });
    sender.send(buffer).then(() => {
      reply.raw.write("event: connected\n");
      reply.raw.write(`data: ${jobId}\n`);
      reply.raw.write("id: 0\n\n");
    });
    clients[jobId] = reply;
  }
);

/* Database */

/* Création du répertoire de la base de données s'il n'existe pas */
createDbDirectory();

/* Ouvrir la base de données */
const db = openDatabase();

/* Créer les tables si elles n'existent pas */
createTables(db);

/* Route pour créer un utilisateur */
fastify.post<{
  Body: {
    login: string;
    password: string;
    permissions: number;
  };
}>("/users", async (request, reply) => {
  const { login, password, permissions } = request.body;
  insertUser(db, login, password, permissions);
  reply.send({ success: true });
});

/* Route pour mettre à jour le login d'un utilisateur */
fastify.put<{
  Params: { id: string };
  Body: { newLogin: string };
}>("/users/:id/login", async (request, reply) => {
  const { id } = request.params;
  const { newLogin } = request.body;
  updateUserLogin(db, parseInt(id), newLogin);
  reply.send({ success: true });
});

/* Route pour mettre à jour le mot de passe d'un utilisateur */
fastify.put<{
  Params: { id: string };
  Body: { newPassword: string };
}>("/users/:id/password", async (request, reply) => {
  const { id } = request.params;
  const { newPassword } = request.body;
  await updateUserPassword(db, parseInt(id), newPassword);
  reply.send({ success: true });
});

/* Route pour mettre à jour les permissions d'un utilisateur */
fastify.put<{
  Params: { id: string };
  Body: { newPermissions: number };
}>("/users/:id/permissions", async (request, reply) => {
  const { id } = request.params;
  const { newPermissions } = request.body;
  await updateUserPermissions(db, parseInt(id), newPermissions);
  reply.send({ success: true });
});

/* Route pour supprimer un utilisateur par son ID */
fastify.delete<{
  Params: { id: string };
}>("/users/:id", async (request, reply) => {
  const { id } = request.params;
  await deleteUserById(db, parseInt(id));
  reply.send({ success: true });
});

/* Route pour supprimer un utilisateur par son login */
fastify.delete<{
  Params: { login: string };
}>("/users/login/:login", async (request, reply) => {
  const { login } = request.params;
  await deleteUserByLogin(db, login);
  reply.send({ success: true });
});

/* Route pour supprimer tous les utilisateurs */
fastify.delete("/users", async (request, reply) => {
  await deleteAllUsers(db);
  reply.send({ success: true });
});

/* Route pour récupérer tous les utilisateurs */
fastify.get("/users", async (request, reply) => {
  const users = await selectAllUsers(db);
  reply.send(users);
});

/* Route pour récupérer un utilisateur par son ID */
fastify.get<{
  Params: { id: string };
}>("/users/:id", async (request, reply) => {
  const { id } = request.params;
  const user = await selectUserById(db, parseInt(id));
  reply.send(user);
});

/* Route pour récupérer un utilisateur par son login */
fastify.get<{
  Params: { login: string };
}>("/users/login/:login", async (request, reply) => {
  const { login } = request.params;
  const user = await selectUserByLogin(db, login);
  reply.send(user);
});

/* Route pour créer un language */
fastify.post<{
  Body: {
    designation: string;
    version: string;
  };
}>("/languages", async (request, reply) => {
  const { designation, version } = request.body;
  insertLanguage(db, designation, parseInt(version));
  reply.send({ success: true });
});

/* Route pour mettre à jour la désignation d'un language */
fastify.put<{
  Params: { id: string };
  Body: { newDesignation: string };
}>("/languages/:id/designation", async (request, reply) => {
  const { id } = request.params;
  const { newDesignation } = request.body;
  updateLanguageDesignation(db, parseInt(id), newDesignation);
  reply.send({ success: true });
});

/* Route pour mettre à jour la version d'un language */
fastify.put<{
  Params: { id: string };
  Body: { newVersion: number };
}>("/languages/:id/version", async (request, reply) => {
  const { id } = request.params;
  const { newVersion } = request.body;
  updateLanguageVersion(db, parseInt(id), newVersion);
  reply.send({ success: true });
});

/* Route pour supprimer un language */
fastify.delete<{
  Params: { id: string };
}>("/languages/:id", async (request, reply) => {
  const { id } = request.params;
  deleteLanguage(db, parseInt(id));
  reply.send({ success: true });
});

/* Route pour supprimer tous les languages */
fastify.delete("/languages", async (request, reply) => {
  deleteAllLanguages(db);
  reply.send({ success: true });
});

/* Route pour récupérer un language par son ID */
fastify.get<{
  Params: { id: string };
}>("/languages/:id", async (request, reply) => {
  const { id } = request.params;
  const language = await selectLanguageById(db, parseInt(id));
  reply.send(language);
});

/* Route pour récupérer tous les languages */
fastify.get("/languages", async (request, reply) => {
  const languages = await selectAllLanguages(db);
  reply.send(languages);
});

/* Route pour créer un work */
fastify.post<{
  Body: {
    id_user: number;
    link: string;
    id_language: number;
    code: string;
  };
}>("/works", async (request, reply) => {
  const { id_user, link, id_language, code } = request.body;
  insertWork(db, link, id_user, id_language, code);
  reply.send({ success: true });
});

/* Route pour récupérer tous les works */
fastify.get("/works", async (request, reply) => {
  const works = await selectAllWorks(db);
  reply.send(works);
});

/* Route pour supprimer tous les works */
fastify.delete("/works", async (request, reply) => {
  deleteAllWorks(db);
  reply.send({ success: true });
});

/* Route pour supprimer un work par son ID */
fastify.delete<{
  Params: { id: string };
}>("/works/:id", async (request, reply) => {
  const { id } = request.params;
  deleteWork(db, parseInt(id));
  reply.send({ success: true });
});

/* Route pour récupérer un work par son ID */
fastify.get<{
  Params: { id: string };
}>("/works/:id", async (request, reply) => {
  const { id } = request.params;
  const work = await selectWorkById(db, parseInt(id));
  reply.send(work);
});

/* Forward output est une fonction asynchrone qui permet de récupérer les messages envoyés par le container et de les renvoyer au client */
async function forwardOutput() {
  for await (const [buff] of receiver) {
    const messageType = buff.readInt8();
    const jobId = buff.subarray(1, 33).toString();
    const reply = clients[jobId];
    if (!reply) {
      continue;
    }
    console.debug(`Forwarding message type ${messageType} for job ${jobId}`);
    const raw = reply.raw;
    switch (messageType) {
      case 1:
      case 2:
        const text = encodeURIComponent(buff.subarray(33).toString());
        raw.write("event: message\n");
        if (messageType === 1) {
          raw.write("id: stdout\n");
        } else {
          raw.write("id: stderr\n");
        }
        raw.write(`data: ${text}\n\n`);
        break;
      case 3:
        const exitCode = buff.readUint32BE(33);
        raw.write("event: message\n");
        raw.write("id: exit\n");
        raw.write(`data: ${exitCode}\n\n`);
        raw.end();
        break;
      default:
        console.error("Unknown message type", messageType);
    }
  }
}

/* Lancer le serveur et la fonction forwardOutput sur le même thread en parallèle */
await Promise.all([fastify.listen({ port: 3000 }), forwardOutput()]);
